
//////////////////////////////////////////////////////////////////////
//
//  XBPGETC.PRG
//
//  Copyright:
//       Alaska Software, (c) 1999-2019. All rights reserved.
//
//  Contents:
//       The XbpGetController class performs pre- and postvalidation of XbpGet
//       objects and navigates between input controls in a window. The usage is
//       as follows:
//
//         oWindow is a window containing XbpGet objects
//
//         // Create a controller object
//         oController := XbpGetController():new( oWindow, lReadExit, lDlgConfirm, lAddControls ):create()
//                        [<lReadExit>]    -> <lToggle> determines wheter the command READ or the
//                                            function ReadModal() can be terminated by the up arrow
//                                            and down arrow keys typed in the first or last valid get.
//                                            .t. -> undo the changes and exit the Get System without
//                                                   warning
//                                            .f. -> lets the application take the decisions like:
//                                                   . remains in the current get
//                                                   . undo the data and remains the current Get
//                                                   . check if all entered fields are valid and
//                                                     if all they are valid close the Get System or
//                                                     set focus to the invalid Get (oControl:close())
//                                             Default: status atual de ReadExit()
//                        [<lDlgConfirm>]  -> confirming the entry data at the end of the dialog?
//                                            Default .f.
//                        [<lAddControls>] -> incluir para navegao Xbps diferentes de XbpGet()
//                                            (ex.: XbpPushButton())?
//                                            Assume .f. (para compatibilidade com os cdigos usados no POINTER)
//
//         // 2nd variant: a modeless Read requires an event Loop
//         oController:read( , .F. )
//         DO WHILE Condition
//            nEvent := AppEvent( mp1, mp2, oXbp )
//            oXbp:handleEvent( nEvent, mp1, mp2 )
//         ENDDO
//
//  Remarks:
//       See the function CabRead_() in CabGetX.prg for detail of the end of the Get System
   /*
    * Terminate READ when application window is closed
    *
    * NOTE:
    *
    * when the XbpGetController object is associated with a window, the
    * window does not need to call :killRead(nExitState) in its :close code block.
    * We must do this here, because XbpGetController() does not know SetAppWindow(),
    * and clicking the X-button in the application window must
    * terminate the application. ::killRead(nExitState) terminates the event
    * loop running in oCtrl:read()
    */
//
//       The :create() method replaces all callback code blocks required for
//       focus change and keyboard navigation.
//
//////////////////////////////////////////////////////////////////////


#include 'MSI.ch'

#define  CB_CLOSE                 1
#define  CB_SETINPUTFOCUS         2
#define  CB_SETDISPLAYFOCUS       3
#define  CB_KILLDISPLAYFOCUS      4
#define  CB_KEYBOARD              5
#define  CB_TABACTIVATE           6
#define  CB_GETSET_COUNT            6

#define  SV_GET_CB                1
#define  SV_XBP                   2
#define  SV_OLD_CB                3
#define  SV_COUNT                   3


CLASS XbpGetController
   PROTECTED:
   VAR oPushBarr

   VAR aControls            // Collected controls in ::collect()
   VAR aCallback            // CB_...   (Call Back)
   VAR nLenControls
   VAR aGetSet
   VAR aNewCB               // CB (Call Back))
   VAR nCurrent
   VAR oWindow
   VAR nModalState
   VAR hasDisplayFocus

   VAR lReadExit            // Up and down keys finish the edition?
   VAR lDlgConfirm          // Ask for confirmation of the data entered?
   VAR lAddControls         // Add for navigation other controls than XbpGet()?
   VAR readExit

   METHOD collect
   METHOD saveCallBack
   METHOD getCallback
   METHOD restCallBack

   EXPORTED:
                            ** These must be exported
   VAR aGetList  READONLY    // Has the editable Xbps
   VAR thisXbp   READONLY    // Current editing control
   VAR lastXbp   READONLY    // Last    editing control
   VAR nextXbp   READONLY    // Next    editing control that will receive input focus
   VAR exitState READONLY    // Last exit state
   VAR doPostValidate        // Enable validation

   METHOD init, create, destroy

                             ** Methods reserved for XbpDialog/XbpCrt windows
   METHOD setDisplayFocus    // window obtains display focus
   METHOD killDisplayFocus   // window looses display focus
   METHOD setInputFocus      // window obtains input focus
   METHOD close              // window is about to be closed

                             ** Methods reserved for other XBPs
   METHOD addEditControl     // Adds an editable XBP
   METHOD addFocusControl    // Adds an XBP that can get input focus
   METHOD tabActivate        // intercepts Tab activation events
   METHOD keyboard           // keyboard navigation for XBPs contained in window

                             ** Methods for navigation
   METHOD first              // go to first control
   METHOD last               // go to last control
   METHOD up                 // go to previous control
   METHOD down               // go to next control

   INLINE METHOD activate(oXbp, nFound)            // <nFound> | <oXbp>
          if nFound     <> NIL
             ::nCurrent := nFound
          else
             ::nCurrent := AScan(::aControls, oXbp)
          endif

          ::exitState := GE_UP            // Para provocar :undo() na XbpGet() que tem o foco
          SetAppFocus(::aControls[::nCurrent])
          ::clearEventLoop()
   RETURN oXbp

                             ** Methods for editing
   METHOD preValidate        // data validation on single entry field level
   METHOD postValidate
   METHOD checkAll           // checks if all XbpGet is valid
   METHOD clearEventLoop     // clear event loop after validation
   METHOD read               // start editing
   METHOD killRead           // enforce edit termination
   METHOD getActive          // returns current edit control
   METHOD getList            // array containing edit controls
   METHOD setData            // transfer data into edit controls
   METHOD getData            // get data from edit controls
   METHOD undo               // undo all XbpGet:changed
   METHOD findGet            // locate a control by its name in edit controls
ENDCLASS


METHOD XbpGetController:init( oWindow, lReadExit, lDlgConfirm, lAddControls )
   ::oWindow         := oWindow
   ::nModalState     := oWindow:getModalState()
   if lReadExit      == NIL
      ::lReadExit    := ReadExit()
   else
      ::lReadExit    := lReadExit
   endif
   if lDlgConfirm    == NIL
      ::lDlgConfirm  := .f.
   else
      ::lDlgConfirm  := lDlgConfirm
   endif
   if lAddControls   == NIL
      ::lAddControls := .f.
   else
      ::lAddControls := lAddControls
   endif

   ::aControls       := {}
   ::aGetList        := {}
   ::aCallback       := {}
   ::readExit        := .f.
   ::hasDisplayFocus := .F.
   ::lastXbp         := NIL
   ::thisXbp         := NIL
   ::nextXbp         := NIL
   ::nCurrent        := 1
   ::exitState       := GE_NOEXIT
   ::dopostValidate  := .t.
RETURN self


/*
 * Create new callback code blocks and get/set codeblocks to replace
 * existing callback code blocks.
 */
METHOD XbpGetController:create( oWindow )
   STATIC  aPushButtons := {{cabICO_ARROWBLUEFIRST16x16,  {|| PostAppEvent(xbeP_Keyboard, xbeK_CTRL_HOME)}, 'Ir para o incio'},;
                            {cabICO_ARROWBLUEDOWN16x16,   {|| PostAppEvent(xbeP_Keyboard, xbeK_DOWN)},      'Prximo item'},;
                            {cabICO_ARROWBLUEUP16x16,     {|| PostAppEvent(xbeP_Keyboard, xbeK_UP)},        'Item anterior'},;
                            {cabICO_ARROWBLUELAST16x16,   {|| PostAppEvent(xbeP_Keyboard, xbeK_CTRL_END)},  'ltimo item'},;
                            NIL,;
                            {cabICO_DESFAZER16x16,        {|| ::undo()},                                    'Desfazer os dados digitados'},;
                            {cabICO_CHECKVERDE16x16,      {|| ::killRead(GE_BOTTOM)},                       'Conferir e atualizar'},;
                            {cabICO_CLOSERED16x16,        {|| PostAppEvent(xbeP_Keyboard, xbeK_ESC)},       'Encerrar sem atualizar'}}

   LOCAL  nLen

   IF Valtype( oWindow ) == "O"
      ::oWindow := oWindow
   ENDIF
                  //
                  //    Create the pushbuttons' barr
   cabPUSH   PUSHBARR      ::oPushBarr;
             NEW           ::oWindow:drawingarea;
             PUSHBUTTONS   aPushButtons;
             CREATE

   ::aGetSet                      := Array( CB_GETSET_COUNT )
   ::aGetSet[CB_CLOSE           ] := {|o,b| IIf( PCount()<>2, o:close           , o:close            := b ) }
   ::aGetSet[CB_SETINPUTFOCUS   ] := {|o,b| IIf( PCount()<>2, o:setInputFocus   , o:setInputFocus    := b ) }
   ::aGetSet[CB_SETDISPLAYFOCUS ] := {|o,b| IIf( PCount()<>2, o:setDisplayFocus , o:setDisplayFocus  := b ) }
   ::aGetSet[CB_KILLDISPLAYFOCUS] := {|o,b| IIf( PCount()<>2, o:killDisplayFocus, o:killDisplayFocus := b ) }
   ::aGetSet[CB_KEYBOARD        ] := {|o,b| IIf( PCount()<>2, o:keyboard        , o:keyboard         := b ) }
   ::aGetSet[CB_TABACTIVATE     ] := {|o,b| IIf( PCount()<>2, o:tabActivate     , o:tabActivate      := b ) }

   ::aNewCB                       := Array( CB_GETSET_COUNT )
   ::aNewCB [CB_CLOSE           ] := {|mp1,mp2,o| ::close           ( mp1, mp2, o ) }
   ::aNewCB [CB_SETINPUTFOCUS   ] := {|mp1,mp2,o| ::setInputFocus   ( mp1, mp2, o ) }
   ::aNewCB [CB_SETDISPLAYFOCUS ] := {|mp1,mp2,o| ::setDisplayFocus ( mp1, mp2, o ) }
   ::aNewCB [CB_KILLDISPLAYFOCUS] := {|mp1,mp2,o| ::killDisplayFocus( mp1, mp2, o ) }
   ::aNewCB [CB_KEYBOARD        ] := {|mp1,mp2,o| ::keyboard        ( mp1, mp2, o ) }
   ::aNewCB [CB_TABACTIVATE     ] := {|mp1,mp2,o| ::tabActivate     ( mp1, mp2, o ) }

   ::collect(::oPushBarr)        // Para serem os primeiros objetos a serem processados por Self
   nLen := Len(::aControls)
   ::collect( ::oWindow )
   aSize(::aControls, (Len(::aControls) - nLen))

RETURN self


/*
 * Restore the original state in the window and release all arrays.
 */
METHOD XbpGetController:destroy

   /*
    * Remove Xbps that have reference to Self
    */
   ::aGetList := NIL
   AEval( ::aControls, {|o| IIf( IsMemberVar( o, "Controller" ), o:destroy(), NIL ) } )

   /*
    * Restore changed callback codeblocks
    */
   ::restCallback()

   /*
    * Reseting member vars
    */
   ::oPushBarr        := NIL
   ::aControls        := NIL
   ::aCallback        := NIL
   ::aGetSet          := NIL
   ::aNewCB           := NIL
   ::oWindow          := NIL
   ::nLenControls     := NIL
   ::hasDisplayFocus  := NIL
   ::aGetList         := NIL
   ::lastXbp          := NIL
   ::nextXbp          := NIL
   ::lDlgConfirm      := NIL
   ::readExit         := NIL
   ::exitState        := NIL
   RETURN self


/*
 * Save and replace all callback code blocks for keyboard and
 * focus change events. This enables the controller to take corrective
 * action when a pre- or postvalidation fails.
 */
METHOD XbpGetController:collect( oXbp )
   LOCAL aXbp, i, imax

   IF oXbp:isDerivedFrom( "XbpDialog" ) .OR. ;                                                     
      oXbp:isDerivedFrom( "XbpCrt"    )

      ::saveCallback( oXbp, ::aGetSet[CB_CLOSE           ], ::aNewCB[CB_CLOSE           ] )
      ::saveCallback( oXbp, ::aGetSet[CB_SETINPUTFOCUS   ], ::aNewCB[CB_SETINPUTFOCUS   ] )
      ::saveCallback( oXbp, ::aGetSet[CB_KILLDISPLAYFOCUS], ::aNewCB[CB_KILLDISPLAYFOCUS] )
      ::saveCallback( oXbp, ::aGetSet[CB_SETDISPLAYFOCUS ], ::aNewCB[CB_SETDISPLAYFOCUS ] )
   ENDIF

   IF oXbp:isDerivedFrom( "XbpDialog" )
      RETURN ::collect( oXbp:drawingArea )
   ENDIF

   /*
    * ::aGetList collects editable Xbase Parts.
    * ::aControls contains Xbase Parts which can get focus in the window.
    */

   IF IsMemberVar( oXbp, "dataLink" )
      ::addEditControl( oXbp )
      if IsMemberVar( oXbp, "tabStop" ) .and.;
         oXbp:TabStop
         ::addFocusControl( oXbp )
      endif
   ELSE
      IF oXbp:isDerivedFrom( "XbpTabpage" )
         ::saveCallback( oXbp, ::aGetSet[CB_TABACTIVATE], ::aNewCB[CB_TABACTIVATE] )
      ELSEIF IsMemberVar( oXbp, "tabStop" ) .AND. oXbp:TabStop
         ::addFocusControl( oXbp )
      ENDIF

      aXbp := oXbp:childList()
      imax := Len( aXbp )
      FOR i:=1 TO imax
         ::collect( aXbp[i] )
      NEXT
   ENDIF

   IF IsMemberVar( oXbp, "Controller" )
      oXbp:controller        := self
      IF Atail( ::aGetList ) <> oXbp
         AAdd( ::aGetList, oXbp )
      ENDIF
   ENDIF

RETURN self


/*
 * ::aGetList collects editable Xbase Parts.
 */
METHOD XbpGetController:addEditControl( oXbp )
   if     oXbp:isDerivedFrom('XbpGet')
          if oXbp:controller           == NIL .and.;
             AScan( ::aGetList, oXbp ) == 0
             AAdd( ::aGetList, oXbp )
          endif
   elseif ::lAddControls                      .and.;
          AScan( ::aGetList, oXbp ) == 0
          AAdd( ::aGetList, oXbp )
   endif
RETURN self


/*
 * ::aControls contains Xbase Parts which can get focus in the window.
 * The :tabStop flag defines whether or not an XBP is added to that list
 * (keyboard navigation within a window).
 */
METHOD XbpGetController:addFocusControl( oXbp )

   IF     oXbp:isDerivedFrom( "XbpGet" )
        /*
         * Do not set focus to XbpGet() that already is associated to a XbpGetController()
         * This avoids the same XbpGet() to be collected by two or more XbpGetController()
         */
          if oXbp:controller == NIL
             AAdd( ::aControls, oXbp )
        /*
         * XbpGet does not require a :keyboard code block, since its
         * :keyboard() method handles the communication with XbpGetController.
         */
             ::saveCallback( oXbp, ::aGetSet[CB_KEYBOARD], NIL )
          endif
   ELSEIF ::lAddControls
          AAdd( ::aControls, oXbp )
        /*
         * Replace and save the :keyboard code block to ensure that a
         * focus change is handled by XbpGetController.
         */
          ::saveCallback( oXbp, ::aGetSet[CB_KEYBOARD], ::aNewCB[CB_KEYBOARD] )
   endif
RETURN self


/*
 * Save and replace one callback code block.
 */
METHOD XbpGetController:saveCallBack( oXbp, bGetSet, bNew )
   LOCAL bOld := Eval( bGetSet, oXbp )
   LOCAL aSave[ SV_COUNT ]

   aSave[SV_GET_CB] := bGetSet
   aSave[SV_XBP   ] := oXbp
   aSave[SV_OLD_CB] := bOld

   EVal( bGetSet, oXbp, bNew )
   AAdd( ::aCallback, aSave )
RETURN self


/*
 * Return a saved callback code block.
 */
METHOD XbpGetController:getCallback( oXbp, bGetSet )
   LOCAL i := Ascan( ::aCallback, {|a| a[SV_XBP] == oXbp .AND. a[SV_GET_CB] == bGetSet } )
RETURN IIf( i==0, NIL, ::aCallback[i, SV_OLD_CB] )


/*
 * Restore all saved callback code blocks.
 */
METHOD XbpGetController:restCallBack
   AEval( ::aCallback, {|a| Eval( a[SV_GET_CB], a[SV_XBP], a[SV_OLD_CB] ) } )
RETURN self


/*
 * Window gains input focus.
 * Set focus to current editable XBP inside window and enable validation
 */
METHOD XbpGetController:setInputFocus
   LOCAL bBlock

   IF ::oWindow <> NIL .AND. ::oWindow:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED
      /*
       * The window is minimized. Don't set focus.
       */
      RETURN self
   ENDIF
   /*
    * If the asscociated window is a dialog,
    * check whether this event was sent during
    * a restore operation. In this case,
    * internal state may have to be reset to
    * properly reflect the validation state.
    * Otherwise, the user can't switch between
    * tab pages after a restore, for instance.
    */
   IF ::hasDisplayFocus == .F. .AND. ;
      ::oWindow         != NIL .AND. ;
      (::oWindow:IsDerivedFrom(XbpDialog()) == .T. .OR.;
       ::oWindow:IsDerivedFrom(XbpCrt())    == .T.)
      ::setDisplayFocus()

      IF ::exitState == GE_LASTGET
         ::exitState := GE_NOEXIT
      ENDIF

   ENDIF

   ::doPostValidate := .t.

   IF ::lastXbp <> NIL
      ::thisXbp := ::lastXbp
      RETURN SetAppFocus( ::lastXbp )
   ENDIF

   IF ::thisXbp == NIL
      ::thisXbp := ::aControls[ ::nCurrent ]
   ENDIF

   bBlock    := ::getCallBack( ::oWindow, ::aGetSet[ CB_SETINPUTFOCUS ] )
   IF bBlock <> NIL
      /*
       * Evaluate the original :setInputFocus codeblock.
       */
      Eval( bBlock, NIL, NIL, ::oWindow  )
   ENDIF
   SetAppFocus( ::thisXbp )
RETURN self


/*
 * Window gains focus.
 * Set focus to current editable XBP inside window and enable validation
 */
METHOD XbpGetController:setDisplayFocus
   LOCAL bBlock := ::getCallBack( ::oWindow, ::aGetSet[ CB_SETDISPLAYFOCUS ] )

   ::hasDisplayFocus = .T.

   IF bBlock <> NIL
      /*
       * Evaluate the original :setDisplayFocus codeblock.
       */
      Eval( bBlock, NIL, NIL, ::oWindow  )
   ENDIF

RETURN ::setInputFocus()


/*
 * Window looses focus.
 * Keep track of XBP that has focus and disable validation
 */
METHOD XbpGetController:killDisplayFocus
   LOCAL bBlock
   LOCAL oXbp := SetAppFocus()
   LOCAL i    := Ascan( ::aControls, {|o| o == oXbp } )

   ::hasDisplayFocus := .F.

   IF i  > 0           .AND.;
      i <> ::nCurrent
      ::nCurrent := i
      ::thisXbp  := ::aControls[i]
   ENDIF
   
   ::doPostValidate := .f.
   ::lastXbp        := ::thisXbp
   ::thisXbp        := NIL
   ::exitState      := GE_LASTGET

   bBlock    := ::getCallBack( ::oWindow, ::aGetSet[ CB_KILLDISPLAYFOCUS ] )
   IF bBlock <> NIL
      /*
       * Evaluate the original :killDisplayFocus codeblock.
       */
      Eval( bBlock, NIL, NIL, ::oWindow  )
   ENDIF
RETURN self


/*
 * Window is about to be closed.
 */
METHOD XbpGetController:close
   ::killRead(GE_ESCAPE)
RETURN Self


/*
 * A TabPage is clicked.
 */
METHOD XbpGetController:tabActivate( mp1, mp2, oTab )
   LOCAL bBlock := ::getCallback( oTab, ::aGetSet[ CB_TABACTIVATE ] )

   IF ::exitState <> GE_LASTGET
      /*
       * Postvalidation is successful.
       * Evaluate the original :tabActivate code block.
       */
      oTab := Eval( bBlock, mp1, mp2, oTab )
   ELSE
      /*
       * Postvalidation failed.
       * TabPage remains in the foreground.
       */
      oTab := ::thisXbp:setParent()
      DO WHILE ! oTab:isDerivedfrom( "XbpTabpage" )
         oTab := oTab:setParent()
         IF oTab == ::oWindow .OR. oTab:isDerivedFrom( "XbpIWindow" )
            EXIT
         ENDIF
      ENDDO
      oTab:toFront()
   ENDIF
RETURN oTab


/*
 * Keyboard handling
 */
METHOD XbpGetController:keyboard( nKey, mp2, oXbp )
   LOCAL bBlock := SetAppEvent( nKey )
   LOCAL i

   mp2         := NIL                     // Avoiding compiler warning
   ::exitState := GE_NOEXIT

   DO CASE
   CASE bBlock <> NIL
     /*
      * A code block is associated with an event.
      */

   CASE nKey == xbeK_ALT_F4
        ::killRead(GE_ESCAPE)
        RETURN .T.

   CASE nKey == xbeK_ESC
        ::exitState := GE_ESCAPE
        ::killRead(GE_ESCAPE)
        RETURN .t.

   CASE nKey == xbeK_TAB
      bBlock      := {|| ::down(.t.) }

   CASE nKey == xbeK_SH_TAB
      bBlock      := {|| ::up(.t.) }

   CASE ::thisXbp <> NIL .AND.;
        !Empty( ::thisXbp:childList() )
     /*
      * NOTE: XBP contains other controls and must handle the
      *       following keys on its own
      */

   CASE nKey == xbeK_CTRL_HOME
      bBlock      := {|| ::first() }

   CASE nKey == xbeK_CTRL_END
      bBlock      := {|| ::last() }

   CASE nKey == xbeK_ENTER
      IF ::thisXbp:isDerivedFrom( "XbpPushbutton" )
         PostAppEvent( xbeP_Activate, NIL, NIL, ::thisXbp )
      ELSEIF ! ::thisXbp:isDerivedFrom( "XbpMLE" )
         bBlock      := {|| ::down(.f., xbeK_ENTER) }       // Used with ::lReadExit
      ENDIF

   CASE nKey == xbeK_DOWN
      bBlock      := {|| ::down(.f.) }

   CASE nKey == xbeK_UP
      bBlock      := {|| ::up(.f.) }

                  //
                  //    Event posted by XbpGet() when the LbDown button was pressed in <oXbp>
                  //          and XbpGetController() must to check the events
                  //          :killInputFocus() and :setInputFocus()
   CASE nKey == xbeM_LbDown
        bBlock := {|| NIL}

   OTHERWISE

      bBlock      := ::getCallback( ::thisXbp, ::aGetSet[CB_KEYBOARD] )
   ENDCASE

   IF bBlock <> NIL
      i := Ascan( ::aControls, {|o| o == oXbp } )
      IF i > 0 .AND. i <> ::nCurrent
         /*
          * A non-XbpGet object has received input focus via mouse click
          */
         ::nCurrent := i
         ::lastXbp  := ::thisXbp
         ::thisXbp  := ::aControls[i]
      ENDIF
      /*
       * Evaluate the code block associated with a key
       */
      Eval( bBlock, nKey, NIL, ::thisXbp )
   ENDIF
RETURN ( bBlock <> NIL )


/*
 * Validation before a Xbase Part gains input focus.
 *
 */
METHOD XbpGetController:preValidate( oXbp, lIsTab )
   LOCAL lValid

   DEFAULT oXbp TO ::thisXbp

   if     ::lastXbp == oXbp .OR. ::exitState == GE_LASTGET
      /*
       * No prevalidation is required when:
       *  - the window that contains ::lastXbp has lost focus and gets it
       *    it back now. The prevalidation is done already in that case.
       *  - a postvalidation has failed and focus must be set back to
       *    the XbpGet holding illegal data.
       */
          if ::lastXbp == oXbp
             lValid    := ::lastXbp:isEnabled()
          else
             lValid    := oXbp:isEnabled()
          endif   
   elseif !IsMethod( oXbp, "preValidate" )
          lValid := .t.
   else
      /*
       * Focus change between controls in the same window
       * requires a prevalidation.
       */
          if     !oXbp:isVisible()
                 lValid := .f.
          elseif lIsTab        .and.;
                 !oXbp:tabStop
                 lValid := .f.
          elseif IsMemberVar(oXbp, 'editable') .and.;
                 !oXbp:editable
                 lValid := .f.
          else
                 lValid := oXbp:prevalidate()
          endif
   endif

   IF lValid
      IF ::exitState == GE_LASTGET .AND. ::lastXbp <> NIL
         /*
          * Focus goes back to XbpGet where postvalidation failed
          */
         SetAppFocus( ::lastXbp )
         ::exitState := GE_NOEXIT
         ::clearEventLoop()
      ELSE
         ::exitState := GE_NOEXIT
         ::thisXbp   := oXbp
         ::lastXbp   := NIL
         ::nCurrent  := AScan( ::aControls, oXbp )
      ENDIF
   ELSE
      /*
       * Prevalidation failed.
       */
      ::exitState  := GE_WHEN
      if ::lastXbp == NIL
         ::lastXbp := ::thisXbp
         ::thisXbp := NIL
      endif
      SetAppFocus(::lastXbp)
      ::clearEventLoop()
   ENDIF
RETURN lValid


/*
 * Validation before an Xbase Part looses input focus.
 *
 */
METHOD XbpGetController:postValidate( oXbp )
   LOCAL nFound, lValid, lSave

   if ::exitState == GE_ESCAPE .or.;
      ::exitState == GE_TOP
      lValid      := .t.
   else
      if     !::doPostValidate         .or.;
             oXbp        == NIL        .or.;
             ::exitState == GE_WHEN    .or.;
             ::exitState == GE_LASTGET .or.;
             ( ::oWindow:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED .AND. LastAppEvent() <> xbeP_Close )
      /*
       * Postvalidation must return .T. when:
       *  - no XBP has got focus in the window yet (initial display)
       *  - the window containing the XBP looses focus (another window is clicked)
       *  - focus is taken away due to a failed prevalidation
       *  - the window is minimized
       */
             lValid := .t.
      elseif !oXbp:isVisible()
             lValid := .t.
      elseif IsMemberVar(oXbp, 'editable') .and.;
             !oXbp:editable
             lValid := .t.
      elseif oXbp   == ::lastXbp
             lValid := .t.
      elseif !IsMethod(oXbp, 'postvalidate')
             lValid := .t.
      else
             if ::nextXbp          <> NIL
                nFound             := AScan(::aControls, ::nextXbp)
                if     nFound       > ::nCurrent
                       ::exitState := GE_DOWN
                elseif nFound       < ::nCurrent
                       ::exitState := GE_UP
                else
                       ::exitState := GE_NOEXIT
                endif
                ::nextXbp          := NIL
             endif

             lSave            := ::doPostValidate
             ::doPostValidate := .f.
             if oXbp:postValidate()
                lValid      := .t.
                ::lastXbp   := oXbp
             else
                lValid      := .f.
                if     oXbp == ::thisXbp
                       SetAppFocus( oXbp )
                elseif oXbp == ::lastXbp
                       ::thisXbp  := oXbp
                       ::nCurrent := AScan( ::aControls, oXbp )
                       SetAppFocus( oXbp )
                endif
             endif
             ::doPostValidate := lSave
      endif
      if lValid
         ::exitState := GE_WRITE
      else
         ::exitState := GE_NOEXIT
      endif
   endif
RETURN lValid


/*
 * Validation of all XbpGet
 *            RETURN .t. -> all XbpGets are valid
 *                   .f. -> there is a XbpGet invalid
 *                          the focus is set to the invalid XbpGet()
 *
 */
METHOD XbpGetController:checkAll()
LOCAL  nInvalid, nInd, nLen, lOk, oXbp

lOk         := .t.
nLen        := Len(::aControls)
::exitState := GE_DOWN
for nInd    := 1 to nLen
    oXbp    := ::aControls[nInd]

    if     !IsMethod(oXbp, 'preValidate')  .or.;
           !oXbp:isVisible()               .or.;
           !oXbp:isEnabled()
           loop
    elseif IsMemberVar(oXbp, 'editable')   .and.;
           !oXbp:editable                           // oXbp can not to be editable but can to be accessed via <lbDown>and
           loop
    elseif !oXbp:prevalidate()
           loop
    elseif (IsMethod(oXbp, 'postValidate') .and.;
           !oXbp:postValidate())
           lOk      := .f.
           nInvalid := nInd
           exit
    endif
next
if !lOk
   ::nCurrent := nInvalid
endif
RETURN lOk


/*
 * Process pending events
 */
METHOD XbpGetController:clearEventLoop
   LOCAL nEvent, mp1, mp2, oXbp

   DO WHILE NextAppEvent() <> xbe_None
      nEvent               := AppEvent( @mp1, @mp2, @oXbp )
      oXbp:handleEvent( nEvent, mp1, mp2 )
   ENDDO
RETURN self


/*
 * Sets or retrives the current Xbase Part.
 */
METHOD XbpGetController:getActive( oXbp )
   LOCAL  nFound, oOld := IIf( ::thisXbp == NIL, ::lastXbp, ::thisXbp )

   IF oOld == NIL
      oOld := SetAppFocus()
   ENDIF

   IF oXbp      <> NIL
      nFound    := AScan( ::aGetList, oXbp )
      if nFound  > 0
         ::activate(oXbp)
      else
         SetAppFocus(oXbp)
         ::clearEventLoop()
      endif
   ENDIF
RETURN oOld


/*
 * Methods for navigation.
 *
 *         Used only by keyboard events
 */
METHOD XbpGetController:first()
   LOCAL  nCurrent, oNewXbp, nFound

   nCurrent    := ::nCurrent
   ::exitState := GE_UP                   // For undoing the data entered
   for nFound  := 1 to ::nLenControls
       oNewXbp := ::aControls[nFound]
       if (!oNewXbp:isEnabled()                     .or.;   // Avoiding to execute SetAppFocus()
           !oNewXbp:isVisible()                     .or.;
           (IsMemberVar(oNewXbp, 'editable') .and.;
            !oNewXbp:editable))
          loop
       endif
       SetAppFocus(oNewXbp)               // Changes ::exitState
       ::clearEventLoop()
       if ::thisXbp == oNewXbp
          exit
       endif
   next
   if     nFound  > ::nLenControls        // No valid Entry Field
          ApiBip_(BIP_WARNING)
          if !::lReadExit
             SetAppFocus(::oWindow)       // For to keep the screen shown
          else
             ::killRead(GE_TOP)
          endif
   elseif nFound == nCurrent
          ApiBip_(BIP_WARNING)
   endif
RETURN NIL


METHOD XbpGetController:last()
   LOCAL  nCurrent, oNewXbp, nFound

   nCurrent    := ::nCurrent
   ::exitState := GE_LASTGET              // Do not postvalidate. It will to be done in ::killRead()
   for nFound  := ::nLenControls to 1 step -1
       oNewXbp := ::aControls[nFound]
       if (!oNewXbp:isEnabled()                     .or.;   // Avoiding to execute SetAppFocus()
           !oNewXbp:isVisible()                     .or.;
           (IsMemberVar(oNewXbp, 'editable') .and.;
            !oNewXbp:editable))
          loop
       endif
       SetAppFocus(oNewXbp)               // Changes ::exitState
       ::clearEventLoop()
       if ::thisXbp == oNewXbp
          exit
       endif
   next

   if     nFound == 0                     // No valid Entry Field
          ApiBip_(BIP_WARNING)
          if !::lReadExit
             ::First_()
          else
             ::killRead(GE_ESCAPE)
          endif
   elseif nFound == nCurrent
          ApiBip_(BIP_WARNING)
   endif
RETURN NIL


METHOD XbpGetController:Up(lIsTab)
   LOCAL  nCurrent, nFound, oNewXbp

   nCurrent    := ::nCurrent
   ::exitState := GE_UP
   for nFound  := (::nCurrent - 1) to 1 Step -1
       oNewXbp := ::aControls[nFound]
       if (!oNewXbp:isEnabled()                     .or.;   // Avoiding to execute SetAppFocus()
           !oNewXbp:isVisible()                     .or.;
           (IsMemberVar(oNewXbp, 'editable') .and.;
            !oNewXbp:editable)                      .or.;
           (lIsTab                           .and.;
            !oNewXbp:tabStop))
          loop
       endif
       SetAppFocus(oNewXbp)               // Changes ::exitState
       ::clearEventLoop()
       if ::thisXbp == oNewXbp
         exit
       endif
      ::exitState := GE_NOEXIT
   next

   if nFound == 0
      if ::lReadExit
         ::killRead(GE_TOP)
      else
         ApiBip_(BIP_WARNING)
         ::activate(NIL, nCurrent)        // For to keep the screen shown
      endif
   endif
RETURN NIL


METHOD XbpGetController:down(lIsTab, nKey)
   LOCAL  nCurrent, nFound, oNewXbp

   nCurrent    := ::nCurrent
   ::exitState := GE_DOWN
   for nFound  := (::nCurrent + 1) to ::nLenControls
       oNewXbp := ::aControls[nFound]
       if (!oNewXbp:isEnabled()                     .or.;   // Avoiding to execute SetAppFocus()
           !oNewXbp:isVisible()                     .or.;
           (IsMemberVar(oNewXbp, 'editable') .and.;
            !oNewXbp:editable)                      .or.;
           (lIsTab                           .and.;
            !oNewXbp:tabStop))
          loop
       endif
       SetAppFocus(oNewXbp)
       ::clearEventLoop()
       if ::thisXbp == oNewXbp
         exit
      endif
      ::exitState := GE_NOEXIT
   next

   if nFound > ::nLenControls
      if ::lReadExit        .or.;
         nKey == xbeK_ENTER
         ::killRead(GE_BOTTOM)
      else
         ApiBip_(BIP_WARNING)
         ::activate(NIL, nCurrent)
      endif
   endif
RETURN ::thisXbp


/*
 * Modal (default) or modeless read method.
 */
METHOD XbpGetController:read( nPos, lModal )
   LOCAL  nEvent, mp1, mp2, oXbp, oHadFocus

   IF Len(::aGetList) == 0
      RETURN Self
   ENDIF

   lModal := .t.

   ::nLenControls := Len(::aControls)
   if ::oWindow:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED
      ::oWindow:setFrameState( XBPDLG_FRAMESTAT_NORMALIZED )
   endif
   if !::oWindow:isVisible()
      ::oWindow:show()
   endif

   oHadFocus := SetAppFocus(::oWindow)
   ::setData()
   if nPos == NIL            .or.;
      nPos  < 1              .or.;
      nPos  > ::nLenControls
      ::first()
   else
      ::activate(NIL, nPos)
      if !::thisXbp:isEnabled()
         ::first()
      endif
   endif

   if ::oWindow:getModalState() <> XBP_DISP_APPMODAL
      ::oWindow:setModalState(XBP_DISP_APPMODAL)
   endif
                  //
                  //    Own event loop
   DO WHILE !::readExit
      nEvent := AppEvent( @mp1, @mp2, @oXbp )
                  //
                  //    ::postValidate() is activated by ::thisXbp:kilInputfocus() that occurs before
                  //    <oXbp> gains input focus and must to know if <oXbp> is after or before ::thisXbp
      if nEvent    == xbeM_LbDown
         ::nextXbp := oXbp
      endif
      oXbp:handleEvent(nEvent, mp1, mp2)
   ENDDO

   if ::oWindow:getModalState() <> ::nModalState
      ::oWindow:setModalState(::nModalState)
   endif
   SetAppFocus(oHadFocus)
RETURN self


/*
 * The :close() method performs a postvalidation. If this is successful,
 * :readExit is set to .T. and the event loop in :read() is exited.
 */
METHOD XbpGetController:killRead(nExitState)
   LOCAL  bBlock
   do while .t.
      ::exitState    := nExitState
      if ::exitState == GE_ESCAPE .or.;
         ::exitState == GE_TOP
                  //
                  //    The validation will be done in ::CheckAll()
         ::thisXbp   := NIL
         ::lastXbp   := NIL
         ::undo()
         ::readExit  := .t.
         exit
      endif

   /*
    * Check if all XbpGet() are valid
    */
      if !::checkAll()
         if !::lDlgConfirm
            ::activate(::aControls[::nCurrent])
            ::readExit := .f.
            exit
         endif

         if (ConfirmBox(NIL, 'ABANDONAR AS ALTERAES FEITAS?',;
                             'CANCELAR AS ATUALIZAES',;
                             XBPMB_OKCANCEL,;
                             (XBPMB_QUESTION + XBPMB_SYSMODAL + XBPMB_MOVEABLE),;
                             XBPMB_DEFBUTTON2) == XBPMB_RET_OK)
            ::undo()
            ::exitState := GE_ESCAPE
            ::readExit  := .t.
         else
            ::activate(::aControls[::nCurrent])
            ::readExit := .f.
         endif
         exit
      endif

      if !::lDlgConfirm
         ::getData(.f.)          // Update the edited values
         ::exitState := GE_WRITE
         ::readExit  := .t.
         exit
      endif

      if (ConfirmBox(NIL, 'CONFIRMA as modificaes feitas?',;
                          'CONFIRMAO DOS DADOS',;
                          XBPMB_OKCANCEL,;
                          (XBPMB_QUESTION + XBPMB_SYSMODAL + XBPMB_MOVEABLE),;
                          XBPMB_DEFBUTTON1) == XBPMB_RET_OK)
         ::getData(.f.)          // Update the edited values
         ::exitState := GE_WRITE
         ::readExit  := .t.
      else
         ::first()
         if ::exitState == GE_TOP         // There is no valid entry field
            if ::lDlgConfirm
               ConfirmBox(NIL, 'NO FORAM ENCONTRADOS dados para edio',;
                               'PROCEDIMENTO ENCERRADO',;
                               XBPMB_OK,;
                               (XBPMB_CRITICAL + XBPMB_SYSMODAL + XBPMB_MOVEABLE))
            endif
            ::readExit := .t.
         else
            ::readExit := .f.
         endif
      endif
      exit
   enddo

   if ::readExit
            /*
             * Window can be closed. Evaluate the original :close code block
             */
      SetAppFocus(::oWindow)
      if ::oWindow:getModalState() <> ::nModalState
         ::oWindow:setModalState(::nModalState)
      endif
      bBlock    := ::getCallback( ::oWindow, ::aGetSet[ CB_CLOSE ] )
      IF bBlock <> NIL
         Eval( bBlock, NIL, NIL, ::oWindow  )
      ENDIF
   endif
RETURN ::readExit


METHOD XbpGetController:getList
RETURN AClone( ::aGetList )


METHOD XbpGetController:setData
   LOCAL  nInd, nLen, oXbp
   nLen := Len(::aGetList)
   for nInd := 1 to nLen
       oXbp := ::aGetList[nInd]
       if !oXbp:isDerivedFrom('xbpget') .and.;     // XbpGet executes XbpGet:setData() when is created
          oXbp:dataLink <> NIL
          oXbp:SetData()
       endif
   next
RETURN self


METHOD XbpGetController:getData(lSetOriginal)
   AEval( ::aGetlist, {|o| o:getData(lSetOriginal) } )
RETURN self


METHOD XbpGetController:undo()
   AEval(::aGetlist, {|o| iif((IsMethod(o, 'undo')          .and.;
                                o:changed), o:undo(), NIL)})
RETURN Self

METHOD XbpGetController:findGet(cGetName)
LOCAL  nFind, oGet

if !Empty(cGetName)          .and.;
   ValType(cGetName) == 'C'
   cGetName          := Lower(cGetName)
   nFind             := AsCan(::aGetList, {|x| (IsmemberVar(x, 'name') .and. Lower(x:name) == cGetName)})
   if nFind           > 0
      oGet           := ::aGetList[nFind]
   endif
endif
RETURN oGet